package CFITSIO::Simple;

###########################################################################
#
# CFITSIO::Simple
#
# Provide a simple, consistent way of reading and writing binary FITS tables.
#
#  fits_read_hdr:     Read the header keywords from a FITS HDU
#  fits_read_bintbl:  Read all or part of a FITS binary table
#  fits_write_bintbl: Write a binary table into a new or existing FITS file
#
###########################################################################

require 5.005_62;
use strict;
use warnings;

use Carp;
use PDL;
use Astro::FITS::CFITSIO qw( :longnames);
use POSIX qw(strtol strtod);

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use CFITSIO::Simple ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( );

our @EXPORT_OK = ( );

our @EXPORT = qw(
		 fits_read_hdr
		 fits_read_bintbl
		 fits_write_bintbl	
		 );
our $VERSION = '0.09';


#  Preloaded methods go here.

our ($ttype, $cfitsio_type, $firstrow);
our (%cfitsio_to_tform, %tform_to_cfitsio, %c_to_tform);
init_conversion_maps();

1;

##*******************************************************************
sub fits_read_bintbl {
##*******************************************************************
# Simple wrapper around Chandra::Tools::Common::read_bintbl_cols()
# to make it more like mrdfits...
    local $_;

    my $file = shift || do {
	carp "fits_read_bintbl: Need filename as first argument\n";
	return ();
    };

    # Pull any options (in a hash reference) off the end

    my %par = ();
    if (ref $_[-1] eq 'HASH') {
	%par = %{pop @_};
    }

    # Specify extension in the file name unless already done

    $file .= $par{extname} ? "[$par{extname}]" : '[1]' 
	unless ($file =~ /\[.+\]/);

    # Make sure read_bintbl_cols doesn't try to move to an extension
    delete $par{extname};	

    # Remaining args (optional) are col names
    my @col_names = @_;		
    
    # Call read_bintbl_cols to actually read the data.  

    return read_bintbl_cols($file, @col_names, {rethash => 1,
					        %par} );
}

##*******************************************************************
sub fits_read_hdr {
##*******************************************************************
# Modified version of P. Ratzlaff's fits_read_header which allows
# for reading keywords from arbitrary extension
    local $_;
    my $RE_Float = qr/[+-]?(?:\d+[.]?\d*|[.]\d+)(?:[dDeE][+-]?\d+)?/;
    my $RE_Long  = qr/[+-]?\d+/;

    unless (@_) {
	carp 'Usage: fits_read_header(file|fitsfilePtr,[extension])';
	return;
    }

    my ($fitsfile, $status);
    my ($n, $left, %header, $key, $value, $comment, @header_list, $type);

    # Read the arguments.  HDU is optional
    my %par = (list => 0,
	       (ref $_[-1] eq 'HASH') ? %{pop @_} : ());

    my $file = shift;
    my $hdu  = shift;

    my $obj_passed = 0; # were we passed a fitsfilePtr?

    $status = 0;
    if (UNIVERSAL::isa($file,'fitsfilePtr')) {
	$fitsfile = $file;
	$obj_passed = 1;
    } else {
	# Open the file.
	fits_open_file($fitsfile, $file, Astro::FITS::CFITSIO::READONLY(), $status);
    }

    # Now we have an open file -- check that status is good before
    # proceeding
    unless ($status) {

	# Go to proper HDU if needed.  Don't use HDU if one was specified in
	# the file name

	if ($hdu && $file !~ /\[.+\]/) {
	    if ($hdu =~ /^\d+$/) { # only digits, so its an extension number
		fits_movabs_hdu($fitsfile, $hdu, undef, $status);
	    } else {		 # Extension name
		fits_movnam_hdu($fitsfile, Astro::FITS::CFITSIO::ANY_HDU(), $hdu, 0, $status);
	    }
	    fits_report_error(*STDERR, $status);
	    return () if ($status);
	}

	# Get the number of fits keywords in primary header
	$fitsfile->get_hdrspace($n, $left, $status);

	# Loop over the keys
	for my $i (1..$n) {
	    last unless $status == 0;

	    $fitsfile->read_keyn($i, $key, $value, $comment, $status);

	    # Guess the type of the keyword
	    $value =~ /\A('?)([^']*)('?)\Z/;
	    $value = $2;
	    if ($1 and $3) {
		$type = 'string';
	    } else {
		if ($value =~ /\A$RE_Long\Z/) { $type = 'long' }
		elsif ($value =~ /\A$RE_Float\Z/) { $type = 'double' }
		elsif ($value =~ /\A([tf])/i) {
		    $type = 'logical';
		    $value = ($1 =~ /t/i) ? 1 : 0;
		}
		else { $type = 'string' }
	    }

	    if ($key eq 'CONTINUE') {
		$header_list[-1]->{val} =~ s/\&\s*\Z//;
		$comment =~ /\A\s*('?)([^']*)('?)/;
		$header_list[-1]->{val} .= $2 if $2;
	    } else {
		push @header_list, {row => $i,
				    key => $key,
				    val => $value,
				    type => $type,
				    comment => $comment};
	    }

	    # Store the key/value in a hash
	    $header{$key} = $value;

	    # Store the comments.
	    if (! exists $header{COMMENTS}{$key}) {
		$header{COMMENTS}{$key} = $comment;
	    }
	    # HISTORY keywords, for instance, can be numerous
	    else {
		if (! ref $header{COMMENTS}{$key}) {
		    $header{COMMENTS}{$key} = [ $header{COMMENTS}{$key} ];
		}
		push @{$header{COMMENTS}{$key}}, $comment;
	    }

	}

	# Close the file if we opened it
	$fitsfile->close_file($status) unless $obj_passed;
    }

    # Report an error - may not always want to write to STDERR...
    fits_report_error(*STDERR, $status);

    # Return either the hash of keywords or the header list
    my $return = $par{list} ? [ @header_list ] : { %header };
    return wantarray ? ($return, $status) : $return;
}

#################################################################################
sub fits_write_bintbl {
#################################################################################
    my $sub = 'fits_write_bintbl';
    local $_;

    my ($name, $status);
    my ($firstelem, $i_col, $col_of_data);
    my $fptr = undef;
    my $file = undef;

    my $file_or_fptr = shift || do {
	carp "$sub: Need a file name or FITS file pointer\n";
	return ();
    };

    (UNIVERSAL::isa($file_or_fptr,'fitsfilePtr') ? $fptr : $file) = $file_or_fptr;

    # Define default parameters, and append any params which appear as
    # a hash ref at the end of the input argument list

    my %par = (extname => 'BINTBL',
	       clobber => 1,
	       close   => 1,);
    if (ref $_[-1] eq 'HASH') {
	%par = (%par,
		%{pop @_});
    }

    # Remaining input elements the data hash, which is required unless a file
    # pointer was supplied

    my %data = @_;
    unless ($fptr || %data) {
	carp "$sub: Need some data\n";
	return ();
    };

    # Check data type for each input element and set up TFORM (format),
    # TTYPE (name), TUNIT (units), cfitsio_type

    if ($file) {
	($fptr, $ttype, $cfitsio_type) =  init_file_for_writing($file, \%par, %data);
	$firstrow  = 1;  # /* first row in table to write   */
    }
   
    # Find the number and maximum number of rows in input data
    my $max_n_rows = get_max_rows(%data);

    # Write the data
    my $ninc;
    my $row0 = 0;
    my %nrow;

    # Determine number of rows for each col
    foreach (keys %data) {
	$nrow{$_} = UNIVERSAL::isa($data{$_},'PDL') ? ($data{$_}->dims())[-1] : @{$data{$_}};
    }

    $status=0;
    fits_get_rowsize($fptr, $ninc, $status);  # Set $ninc, the optimal number of rows to write
    while ($row0 < $max_n_rows) {
	for $i_col (0 .. $#{$ttype}) {
	    $name = $ttype->[$i_col];
	    next unless (exists $data{$name});

	    my $row1 = $row0 + $ninc;
	    $row1 = $nrow{$name}-1 if $row1 >= $nrow{$name};
	    next if ($row1 < $row0);

	    if (UNIVERSAL::isa($data{$name},'PDL')) {
		my $dims = $data{$name}->dims - 1;
		my $slice = ":," x $dims . "$row0:$row1";
		$col_of_data = [ list $data{$name}->slice($slice) ];
	    } else {
		$col_of_data = [ @{$data{$name}}[$row0..$row1] ];
	    }

	    fits_write_col($fptr, $cfitsio_type->{$name}, $i_col+1, $firstrow + $row0, 1,
			   scalar @{$col_of_data}, $col_of_data, $status);
	    return if ($status && fits_error($status));
	}
	$row0 += $ninc;
    }

    # Close file if needed

    if ($par{close}) {
	fits_write_chksum($fptr,$status);
	fits_close_file($fptr, $status);
	return if ($status && fits_error($status));
    }

    $firstrow += $max_n_rows;
    return wantarray ? (1, $fptr) : 1;
}


##########################################################################
sub get_max_rows {
##########################################################################
    my (%data) = @_;

    my $max_n_rows = 0;
    my @naxes;
    my $n_rows;

    foreach $_ (keys %data) {
	if (UNIVERSAL::isa($data{$_},'PDL')) {
	    $n_rows = ($data{$_}->dims())[-1];
	} else {
	    $n_rows = scalar @{$data{$_}};
	}
	$max_n_rows = $n_rows if ($n_rows > $max_n_rows);
    }
    
    return $max_n_rows;
}
	    


##########################################################################
sub init_file_for_writing {
##########################################################################
    my $sub = 'init_file_for_writing';

    my ($file, $par, %data) = @_;

    my ($hdunum, $hdutype, $n_rows, $keyword, $i_col, @naxes, $name);
    my ($key, $val, $type, $comment);
    my $status = 0;

    my $ttype = [];
    my $tform = [];
    my $tunit = [];
    my @dims  = ();
    my @descr = ();
    my $tfields;
    my $cfitsio_type = {};

    my %par = %{$par};
    my $fptr;

    # Figure out the ordered list of columns to write.  First type col_names parameter, then
    # try colkeys (from a previous file read), finally just sort the input data keywords

    my @col_names;
    if ($par{col_names}) {
	@col_names = @{$par{col_names}};
    } elsif ($par{colkeys}) {
	@col_names = sort { $par{colkeys}->{$a}->{index} <=> $par{colkeys}->{$b}->{index} } keys %{$par{colkeys}};
    } else {
	@col_names = sort keys %data;
    }
    my $n_cols = scalar @col_names;

    # Setup input for CFITSIO fits_create_table routine by looking at data for
    # each column

    foreach $name (@col_names) {
	if (UNIVERSAL::isa($data{$name},'PDL')) {

	    # Get the PDL data type and map it into the CFITSIO type
	    unless ($cfitsio_type->{$name} = match_datatype($data{$name})) {
		carp "$sub: No CFITSIO type matches PDL type for col $name\n";
		return ();
	    }

	    my $n_elem = 1;
	    my @naxes = $data{$name}->dims();
	    map { $n_elem *= $naxes[$_] } (0 .. $#naxes-1);

	    push @{$tform}, $par{colkeys}->{$name}->{tform} || $n_elem . $cfitsio_to_tform{$cfitsio_type->{$name}};
	    push @{$ttype}, $name;
	    push @descr, $par{colkeys}->{$name}->{descr} || '';
	    push @{$tunit}, $par{units}->{$name} || $par{colkeys}->{$name}->{tunit} || '';
	    push @dims, [ @naxes ];
	} else {   # It's a perl list

	    # For lists, either use the user-specified TFORM, or else scan
	    # through data and pick one of LONG, DOUBLE, or STRING

	    my $tform1 =  get_tform(uc $par{tform}->{$name}, $data{$name});
	    my $tform_char;
	    ($tform_char = $tform1) =~ s/\d//g;

	    push @{$tform}, ($tform1 =~ /\d/) ? $tform1 : "1$tform1";
	    push @{$ttype}, $name;
	    push @descr, $par{colkeys}->{$name}->{descr} || '';
	    push @{$tunit}, $par{tunit}->{$name} || $par{colkeys}->{$name}->{tunit} || '';
	    $cfitsio_type->{$name} = $tform_to_cfitsio{$tform_char};
	    push @dims, [ scalar @{$data{$name}} ];
	}
    }

    # If file already exists, then we need to have clobber==1 or
    # append=1 to continue

    if (-e $file) {
	unless ($par{clobber} || $par{append}) {
	    carp "$sub: File '$file' already exists, can't clobber it\n";
	    return ();
	}
	unlink $file unless ($par{append});  # Kill file if not appending
    }

    if (-e $file) {
	# If the file is there, then append bin table extension.  Note that
	# one can specify append=>1 even if there is no existing file

	fits_open_file($fptr, $file, Astro::FITS::CFITSIO::READWRITE(), $status);
	return if ($status && fits_error($status));

	fits_get_num_hdus($fptr, $hdunum, $status);
	return if ($status && fits_error($status));

	fits_movabs_hdu($fptr, $hdunum, $hdutype, $status);
	return if ($status && fits_error($status));
    } else {
	# Otherwise create the file
	
	fits_create_file($fptr, $file,  $status);
	return if ($status && fits_error($status));

	# If the primary_hdu keyword is defined, then copy the primary HDU from that file
	if ($par{primary_hdu}) {
	    my $fptr_in;
	    fits_open_file($fptr_in, $par{primary_hdu}, Astro::FITS::CFITSIO::READONLY(), $status);
	    # copy first HDU of input file to output file
	    fits_copy_file($fptr_in, $fptr, 1, 1, 0, $status);
	    fits_close_file($fptr_in, $status);
	    return if ($status && fits_error($status));
	}
    }

    #    /* append a new empty binary table onto the FITS file */

    $n_rows    = 0;   #  Don't need to know in advance how many rows
    fits_create_tbl($fptr, Astro::FITS::CFITSIO::BINARY_TBL(), $n_rows, $n_cols, $ttype, $tform,
		    $tunit, $par{extname}, $status);
    return if ($status && fits_error($status));

    # Write any FITS header keywords
    
    foreach $keyword (@{$par{keywords}}) {
	# If keyword is a hashref, then parse accordingly.  Otherwise it's a listref
	if (ref $keyword eq 'HASH') {
	    # If it's a comment or history entry, write it using the appropriate routine.  Otherwise
	    if ($keyword->{key} eq 'COMMENT') {
		fits_write_comment($fptr, $keyword->{comment} || ' ', $status);
		next;
	    } elsif ($keyword->{key} eq 'HISTORY') {
		fits_write_history($fptr, $keyword->{comment} || ' ', $status);
		next;
	    } else {
		($key, $val, $type, $comment) = 
		  ($keyword->{key}, $keyword->{val}, $keyword->{type}, $keyword->{comment});
	    }
	} else {
	    ($key, $val, $type, $comment) = @{$keyword};
	}
	next if ($key =~ /^(EXTNAME|XTENSION|BITPIX|TFIELDS|NAXIS|PCOUNT|GCOUNT|CHECKSUM|DATASUM)$/);
	next if ($key =~ /^(TFORM|TTYPE|TUNIT|TDIM|NAXIS)\d+$/);

	$type = $type || 'string';
	$comment = $comment || '';

	# Translate from "C" type to CFITSIO type

	if ($type eq 'string') {
	    fits_write_key_longstr($fptr, $key, $val, $comment, $status);
	} elsif ($type eq 'logical') {
	    fits_write_key_log($fptr, $key, $val, $comment, $status);
	} else {
	    my $tform_type = $c_to_tform{uc($type)};
	    my $keyword_type = $tform_to_cfitsio{$tform_type};
	    fits_write_key ($fptr, $keyword_type, $key, $val, $comment, $status);
	}
    }

    # Fix column descriptions and Set TDIM keyword for any multi-dimensional table elements
    for $i_col (0 .. $n_cols-1) {
	fits_modify_comment($fptr, "TTYPE".($i_col+1), $descr[$i_col], $status)
	  if ($descr[$i_col]);
	
	@naxes = @{$dims[$i_col]};
	pop @naxes;


	if (@naxes > 1) {   # Is is multidimensional?
	    fits_write_tdim($fptr, $i_col+1, scalar @naxes, \@naxes, $status);
	    return if ($status && fits_error($status));
	 }
    }

    return $fptr, $ttype, $cfitsio_type;
}

#################################################################################
sub fits_error {
#################################################################################
    my $status = shift;
    my $errmsg;

    fits_get_errstatus($status, $errmsg);
    carp "Status = $status: $errmsg\n";
    return 1;
}

#################################################################################
sub get_tform {
#################################################################################
    my $tform = shift;
    my $data = shift;
    my $unused_chars;
    my $could_be_long   = 1;
    my $could_be_double = 1;
    my $max_len = 0;
    my ($c, $fits_type);

    # If tform is already defined, then try to convert any "C" data
    # types (e.g. DOUBLE) to the appropriate FITS type, get
    # rid of any whitespace, and return

    if ($tform) {
	$tform =~ s/\s//g;
	while (($c, $fits_type) = each %c_to_tform) {
	    $tform =~ s/^(\d*)$c$/$1$fits_type/;
	}
	return $tform;
    }

    # If tform is not defined, then make a guess based on the POSIX
    # conversion routines.  For a column to be LONG, then every element
    # must successfully convert to LONG.  If not, fall through to trying
    # to convert to DOUBLE.  If that fails, assume it is a STRING.

    foreach (@{$data}) {
	$max_len = length if (length > $max_len);

	if ($could_be_long) {
	    (undef, $unused_chars) = strtol($_);
	    if ($unused_chars == 0) {
		next;
	    } else {
		$could_be_long = 0;
	    }
	}

	if ($could_be_double) {
	    (undef, $unused_chars) = strtod($_);
	    if ($unused_chars == 0) {
		next;
	    } else {
		$could_be_double = 0;
		last;
	    }
	}
    }

    return ('J') if ($could_be_long);
    return ('D') if ($could_be_double);
    return ("${max_len}A");
}

##*******************************************************************
sub init_conversion_maps {
##*******************************************************************
# Define maps between various ways of specifying data types

# Mapping between CFITSIO data type and TFORM letter
%cfitsio_to_tform = (Astro::FITS::CFITSIO::TDOUBLE() => 'D',
	    Astro::FITS::CFITSIO::TFLOAT() => 'E',
	    Astro::FITS::CFITSIO::TSHORT() => 'I',
	    Astro::FITS::CFITSIO::TINT() => 'J',
	    Astro::FITS::CFITSIO::TLONG() => 'J',
	    Astro::FITS::CFITSIO::TBYTE() => 'B',
	    Astro::FITS::CFITSIO::TUSHORT() => 'U',
	    Astro::FITS::CFITSIO::TULONG() => 'V',
	    Astro::FITS::CFITSIO::TSTRING() => 'A',
	      );

# TFORM to CFITSIO
%tform_to_cfitsio = reverse %cfitsio_to_tform;

# Mapping from "C"-ish data types to FITS (TFORM) types.  The "C" types
# can be used by user to specify types (for data and header keywords)
%c_to_tform = (DOUBLE => 'D',
 	    	FLOAT => 'E',
	    	SHORT => 'I',
	    	INT => 'J',
	    	LONG => 'J',
	    	BYTE => 'B',
	    	USHORT => 'U',
	    	ULONG => 'V',
	    	STRING => 'A',
	      );
}

##***************************************************************************
##  Below are several routines which were copied directly from 
##  Chandra::Tools::Common (P.Ratzlaff).  They are included here to 
##  remove dependence on this Chandra-specific package.
##***************************************************************************

sub read_bintbl_cols {

  my @leave_subs_beg = my @leave_subs_end = ();
  my $cleanup = sub { $_->() for @leave_subs_beg, @leave_subs_end; };
  my $add_clean_beg = sub { push @leave_subs_beg, @_ };
  my $add_clean_end = sub { unshift @leave_subs_end, @_ };

  my $old_packing = Astro::FITS::CFITSIO::PerlyUnpacking(-1);
  $add_clean_beg->(sub { Astro::FITS::CFITSIO::PerlyUnpacking($old_packing) }) if $old_packing;

  my %opt = parse_opts(\@_, qw(extname status rfilter ninc dtypes rethash colkeys));

  # first arg is fitsfilePtr or filename
  my $input = shift;
  my $fptr_passed = 0;
  my $fptr;
  my $s = 0;
  if (UNIVERSAL::isa($input,'fitsfilePtr')) {
    $fptr = $input;
    $fptr_passed = 1;
  } else {
    $fptr = Astro::FITS::CFITSIO::open_file($input,Astro::FITS::CFITSIO::READONLY(),$s);

    check_status($s) or
      error("could not open FITS file '$input'"), $cleanup->(), return;
    $add_clean_end->(sub { $fptr->close_file($s=0) } );
  }

  # move to target HDU
  if (exists $opt{extname}){
    my $old_hdu_num;
    $fptr->get_hdu_num($old_hdu_num);
    $fptr->movnam_hdu(Astro::FITS::CFITSIO::ANY_HDU(),$opt{extname},0,$s);
    check_status($s) or
      error("could not move to binary table '$opt{extname}'"), $cleanup->(), return;
    $add_clean_beg->(sub { $fptr->movabs_hdu($old_hdu_num, undef, $s=0) })
      if $fptr_passed;
  }

  my $type;
  $fptr->get_hdu_type($type, $s);

  # make data structure describing the columns available
  my $h = $fptr->read_header;
  my $ncols; $fptr->read_key(Astro::FITS::CFITSIO::TINT(),'tfields',$ncols,undef,$s);
  my %cols;
  for (1..$ncols) {
    my ($name, $format, $dims, $unit, $descr);
    my $naxis = 0;
    my $naxes = [];
    $fptr->read_key(Astro::FITS::CFITSIO::TSTRING(), 'ttype'.$_, $name, $descr, $s);
    $fptr->read_key(Astro::FITS::CFITSIO::TSTRING(), 'tform'.$_, $format, undef, $s);
    $fptr->read_key(Astro::FITS::CFITSIO::TSTRING(), 'tunit'.$_, $unit, undef, $s);
    $s = 0;			# TUNIT is optional, so errors may occur
    $fptr->read_tdim($_, $naxis, $naxes, $s);

    $name = lc $name;
    my ($repeat, $btype) = $format =~ /^(\d*)([A-Z])[\d\.]*$/ or
      error("do not understand format '$format'"), $cleanup->(), return;
    $repeat = 1 unless length $repeat;
    $btype =~ tr/fF/eE/;  # ASCII table can have format='F' for real number
    @{$cols{$name}}{qw( n format repeat btype dims unit descr)}
      = ($_, $format, $repeat, $btype, $naxes, $unit, $descr);
  }

  # read all columns (ordered by their column number), unless argument was given
  my @reqcols = map { lc } @_ ? @_ : sort { $cols{$a}{n} <=> $cols{$b}{n} } keys %cols;

  # Set up to read data into binary data strings
  Astro::FITS::CFITSIO::PerlyUnpacking(0);

  # construct the colkeys output
  my %colkeys;
  if ($opt{colkeys}) {
    my $hdr = $fptr->read_header;
    $hdr->{$_} =~ s/^'//, $hdr->{$_}=~s/\s*'$// for keys %$hdr; # clean string values
    for my $c (@reqcols) {
      %{$colkeys{$c}} = map { (lc($_)=~/(.*?)\d+$/), $hdr->{$_} }
	grep /^t\D+$cols{$c}{n}$/i, keys %$hdr;
      $colkeys{$c}{index} = $cols{$c}{n};
      $colkeys{$c}{descr} = $cols{$c}{descr};
    }
  }

  my %typemap = (
		 'A' => { 'pdl' => undef, 'null_val' => '', },
		 'I' => { 'pdl' => short, 'null_val' => 0, },
		 'J' => { 'pdl' => long, 'null_val' => 0, },
		 'E' => { 'pdl' => float, 'null_val' => 0, },
		 'D' => { 'pdl' => double, 'null_val' => 0, },
		 'X' => { 'pdl' => byte, 'null_val' => 0, },
		 'L' => { 'pdl' => byte, 'null_val' => 0, },
		 'B' => { 'pdl' => byte, 'null_val' => 0, },
		);

  # user cannot specify output types for logical, string and bit columns
  # type specified must be of a PDL::Type token
  croak "dtypes argument must be a hash reference"
    if ($opt{dtypes} and ref $opt{dtypes} ne 'HASH');
  my %user_types = map { lc($_) => $opt{dtypes}{$_} } keys %{$opt{dtypes}};
  for (keys %user_types) {
    if ( $cols{$_}{btype} =~ /^LXA$/i or
	 !UNIVERSAL::isa($user_types{$_},'PDL::Type')
       ) {
      carp("user-specified type for column '$_' being ignored");
      delete $user_types{$_};
    }
  }

  # select datatypes
  for (@reqcols) {
    exists $cols{$_} or
      croak "requested column '$_' not in file";

    exists $typemap{$cols{$_}{btype}} or
      croak "cannot read column type '$cols{$_}{btype}'";

    # the type of piddle we'll read into
    $cols{$_}{ptype} = exists $user_types{$_} ? $user_types{$_} :
      $typemap{$cols{$_}{btype}}{'pdl'};

    # the type we'll tell cfitsio we're reading
    for my $btype ($cols{$_}{btype}) {
      $btype eq 'L' and $cols{$_}{ctype} = Astro::FITS::CFITSIO::TLOGICAL(), last;
      if ($btype eq 'X') {
	  # Read bit field as n_bits/8 bytes
	  $cols{$_}{ctype} = Astro::FITS::CFITSIO::TBYTE();
	  $cols{$_}{repeat} = POSIX::ceil($cols{$_}{repeat} / 8);
	  last;
      }
      $btype eq 'A' and $cols{$_}{ctype} = undef, last;
      $cols{$_}{ctype} = match_datatype($cols{$_}{ptype});
    }
  }

  my $nrows;
  $fptr->get_num_rows($nrows,$s);
  $s = 0;			# what could possibly go wrong?
  if (!$nrows) {
    $cleanup->();
    my @return = $opt{rethash} ?
      map { $_ => $cols{$_}{btype} eq 'A' ? [] : pdl($cols{$_}{ptype}, []) } @reqcols
      :
      map { $cols{$_}{btype} eq 'A' ? [] : pdl($cols{$_}{ptype}, []) } @reqcols;
    unshift @return, \%colkeys if $opt{colkeys};
    return @return;
  }

  my $ninc;
  if (!$opt{ninc}) { $fptr->get_rowsize($ninc, $s) }
  else { $ninc = $opt{ninc} }
  $ninc = $nrows if $nrows < $ninc;

  # create piddles
  for (@reqcols) {
    if ($cols{$_}{btype} ne 'A') {
      for my $r ($cols{$_}{repeat}) {
	$r == 1 and
	  @{$cols{$_}}{qw( pdl tmppdl )} = (
					    zeroes($cols{$_}{ptype}, $nrows),
					    zeroes($cols{$_}{ptype}, $ninc)
					    ), last;
	$r > 1 and
	  @{$cols{$_}}{qw( pdl tmppdl )} = (
					    zeroes($cols{$_}{ptype}, $r, $nrows),
					    zeroes($cols{$_}{ptype}, $r, $ninc)
					    ), last;
	@{$cols{$_}}{qw( pdl tmppdl )} = undef; # repeat of zero is allowed by spec
      }
    }
    else {
      $cols{$_}{'pdl'} = [];
    }
  }

  # create masks if we'll be row filtering
  my ($good_mask, $tmp_good_mask, $ngood);
  if ($opt{rfilter}) {
    $good_mask = ones(byte,$nrows);
    $tmp_good_mask = ones(byte,$ninc);
    $ngood = 0;
  }

  print STDERR ('    ') if ($opt{status});
  my $rows_done = 0;
  while ($rows_done < $nrows) {

    print STDERR (("\b"x4).sprintf(" %2d",int(100*$rows_done/$nrows)).'%') if $opt{status};

    my $rows_this_time = $nrows - $rows_done;
    $rows_this_time = $ninc if $rows_this_time > $ninc;

    # row filter
    if ($opt{rfilter}) {
      my $tmp_ngood = 0;
      $fptr->find_rows($opt{rfilter},$rows_done+1,$rows_this_time,$tmp_ngood,${$tmp_good_mask->get_dataref},$s);
      $tmp_good_mask->upd_data;

      check_status($s) or
	error("error filtering rows: rfilter = '$opt{rfilter}'"),
	  $cleanup->(), return;

      (my $t = $good_mask->slice($rows_done.':'.($rows_done+$rows_this_time-1))) .=
	$tmp_good_mask->slice('0:'.($rows_this_time-1));

      $ngood += $tmp_ngood;

      $tmp_ngood > 0 or
	$rows_done += $rows_this_time,
	  next;
    }

    for (@reqcols) {
      next unless $cols{$_}{repeat};
      if ($cols{$_}{btype} ne 'A') {
	$fptr->read_col( $cols{$_}{ctype},
			 $cols{$_}{n},
			 $rows_done+1, 1,
			 $cols{$_}{repeat} * $rows_this_time,
			 0,
			 ${$cols{$_}{tmppdl}->get_dataref},
			 undef,
			 $s);
	$cols{$_}{tmppdl}->upd_data;

	my $slice1 = ($cols{$_}{repeat} > 1 ? ':,' : '').$rows_done.':'.($rows_done+$rows_this_time-1);
	my $slice2 = ($cols{$_}{repeat} > 1 ? ':,' : '').'0:'.($rows_this_time-1);
	(my $t = $cols{$_}{'pdl'}->slice($slice1)) .=
	  $cols{$_}{tmppdl}->slice($slice2);

      } else {			# string type
	my $tmp = [];
	$fptr->read_col(Astro::FITS::CFITSIO::TSTRING(),
			$cols{$_}{n},
			$rows_done+1, 1,
			$rows_this_time,
			0,
			$tmp,
			undef,
			$s,
		       );
	push @{$cols{$_}{'pdl'}}, @$tmp;
      }

      check_status($s) or
	error("error reading FITS data"), $cleanup->(), return;

    }
    $rows_done += $rows_this_time;
  }

  if ($opt{rfilter}) {
    my $good_index = which($good_mask);

    for (@reqcols) {
      next unless $cols{$_}{repeat};

      if ($cols{$_}{btype} ne 'A') {
	my $r = $cols{$_}{repeat};
	if ($r > 1) {
	  my $index = ($good_index->dummy(0,$r)*$r + sequence($r))->clump(-1);
	  $cols{$_}{'pdl'} = $cols{$_}{'pdl'}
	    ->clump(-1)
	      ->index($index)
		->reshape($r,$good_index->nelem);
	} else {
	  $cols{$_}{'pdl'} = $cols{$_}{'pdl'}->index($good_index);
	}
      } else {			# string type
	@{$cols{$_}{'pdl'}} = @{$cols{$_}{'pdl'}}[$good_index->list];
      }
    }
  }
  print STDERR (("\b"x4).'100%') if ($opt{status});

  $cleanup->();

  # Reshape any non-string columns that have a TDIM<n> keyword indicating
  # sub-dimensions within column
  for (@reqcols) {
      next unless ($cols{$_}{btype} ne 'A' and @{$cols{$_}{dims}} >= 2);
      my @d = dims $cols{$_}{pdl};
      my $size0 = $cols{$_}{pdl}->nelem;
      $cols{$_}{pdl}->reshape(@{$cols{$_}{dims}}, $d[-1]);
      croak "Error - lost data when reshaping column $_ to dims $cols{$_}{dims}\n"
	if ($size0 != $cols{$_}{pdl}->nelem);
  }

  my @retval = $opt{rethash} ?
    map { $_ => $cols{$_}{'pdl'} } @reqcols :
    map { $cols{$_}{'pdl'} } @reqcols;

  unshift @retval, \%colkeys if $opt{colkeys};
  return @retval;
}

# just print an error message to STDOUT
sub error {
  my $txt = shift;
  my $mname = (caller 1)[3];
  defined $mname or $mname = 'main';
  print STDERR "$mname() - $txt\n";
}

sub match_datatype {
  my $arg = shift;

  my $pdl_type;
  if (UNIVERSAL::isa($arg,'PDL')) {
    $pdl_type = $arg->get_datatype;
  } elsif (UNIVERSAL::isa($arg,'PDL::Type')) {
    $pdl_type = $arg->[0];
  } else {
    croak "argument should be a PDL object or PDL::Type token";
  }

  my $pdl_size = PDL::Core::howbig($pdl_type);

  my @cfitsio_possible_types;
  # test for real datatypes
  if ($pdl_type == float(1)->get_datatype or
      $pdl_type == double(1)->get_datatype
     ) {
    @cfitsio_possible_types = (
			       Astro::FITS::CFITSIO::TDOUBLE(),
			       Astro::FITS::CFITSIO::TFLOAT(),
			      );
  } elsif ($pdl_type == short(1)->get_datatype or
	   $pdl_type == long(1)->get_datatype
	  ) {
    @cfitsio_possible_types = (
			       Astro::FITS::CFITSIO::TSHORT(),
			       Astro::FITS::CFITSIO::TINT(),
			       Astro::FITS::CFITSIO::TLONG(),
			      );
  } elsif ($pdl_type == ushort(1)->get_datatype or
	   $pdl_type == byte(1)->get_datatype
	  ) {
    @cfitsio_possible_types = (
			       Astro::FITS::CFITSIO::TBYTE(),
			       Astro::FITS::CFITSIO::TUSHORT(),
			       Astro::FITS::CFITSIO::TUINT(),
			       Astro::FITS::CFITSIO::TULONG(),
			      );
  } else {
    croak "cannot handle PDL type $pdl_type";
  }


  for (@cfitsio_possible_types) {
    return $_ if $pdl_size == Astro::FITS::CFITSIO::sizeof_datatype($_);
  }

  croak "no CFITSIO type for PDL type $pdl_type";
}

sub parse_opts {
  my $args = shift;

  my %opts = ();

  if (ref $args->[-1] eq 'HASH') {

    # copy hash, making keys lowercase as we go
    my $href_in = pop @$args;
    %opts = map { lc($_) => $href_in->{$_} } keys %$href_in;

    # ensure only allowed options are given
    if (@_) {

      # hash for convenient lookups of valid options
      my %allowed;
      @allowed{map lc, @_} = ();

      my $caller = (caller 1)[3] || 'main';
      exists $allowed{$_} or confess $caller."() - option '$_' unknown"
	for keys %opts;
    }
  }

  return %opts;
}

sub check_status {
  my $s = shift;
  if ($s != 0) {
    my $txt;
    Astro::FITS::CFITSIO::fits_get_errstatus($s,$txt);
    carp "CFITSIO error: $txt";
    return 0;
  }

  return 1;
}

__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

CFITSIO::Simple - Simple routines to read and write FITS binary tables

=head1 SYNOPSIS

  use PDL;
  use CFITSIO::Simple;

  $hdr = fits_read_hdr($file, $extension);

  %data = fits_read_bintbl($file, @col_names, {%options});

  $status = fits_write_bintbl($file, %data, {%options});

=head1 DESCRIPTION

B<CFITSIO::Simple> provides a simple, consistent way of reading and writing
FITS binary tables which should suffice for most applications.  It relies on
Astro::FITS::CFITSIO as the underlying FITS file interface.

FITS file extensions can be either a name or number, and can be specified in 
the file name or options without conflict.  An extension in the file name takes
precedence. B<fits_read_bintbl> has the default options are set so
that the minimal call 
 
    %data = fits_read_bintbl($file);
 
does the right thing: return a hash of columns for the first extension.

B<Fits_write_bintbl> will write a new binary table FITS file, or append onto
an existing one.  This routine is a bit more complicated in terms of available
options, but still supports the minimal call which creates a new FITS file with
a single extension containing the column data:

    $status = fits_write_bintbl($file, %data);

=head1 ROUTINES

=over 4

=item fits_read_hdr( )

Usage:

    $hdr = fits_read_hdr($file, $extension)
    ($hdr, $status) = fits_read_hdr($file, $extension, {%options})

Returns a hash reference of all the FITS header keywords in the specified
extension.  The $extension is optional, and will be ignored if an
extension is present in the file name, e.g.

    $file = 'my_table.fits[EVENTS]';
    $file = 'my_table.fits[2]';

If the option {list => 1} is present, then the returned $hdr is an ordered list
of hash refs describing the keywords.  This is useful because it maintains the
order of header keywords in the original file.  The list version of $hdr can be
passed to fits_write_bintbl().  Each hash ref has the keys 'key', 'val', 'row',
and 'comment'.

=item fits_read_bintbl( )

Usage:

    %data = fits_read_bintbl($file, @columns, {%options})

Returns a hash of column data in the binary table.  If a column
is numeric the hash entry will be a PDL object (piddle).  Columns
containing string data are returned as an array reference.
If no column names (@columns) are specified, then all columns
are read.

Available options are are the same as read_bintbl_cols() (see
perldoc for B<Chandra::Tools::Common>).  By default, rethash=>1
and colkeys=>0, but these can be overridden.

=item fits_write_bintbl( )

Usage 1:

  $status = fits_write_bintbl($file, %data, {%options});

Writes a binary table extension, and returns 1 for success and 0 for failure.
The %data hash can be a mix of Perl array refs and PDL objects.
Enough rows are written to accomodate the longest data element.

For PDL objects, the data type is automatically determined.  For Perl array
refs, the user can either explicitly specify the type or let the
routine make its best guess, choosing either LONG, DOUBLE, or STRING.

Usage 2:

  ($status, $fptr) = fits_write_bintbl($file, %data, {close => 0, %options});
  $status = fits_write_bintbl($fptr, %more_data, {close => 0});
  $status = fits_write_bintbl($fptr, {close => 1});

This form writes some data to a binary table, as in Usage 1, but leaves the
file open for subsequent additional write commands into the same table.
This is useful for writing out large files that may not fit into memory.
Subsequent data hashes need not contain all of the original columns.
The final command, with no data supplied, will close the file.  Note
that (close => 1) is the default, so it is not strictly necessary to 
supply this option to close the file.  

Since some information about the FITS file is kept internally in the
package, you cannot write to more than one table at a time.  This
routine should be rewritten to use object methods.

Available options are:

=over 4

=item extname

Name of the binary table extension.  Defaults to 'BINTBL'.

=item clobber

Allow an existing file to be overwritten.  Note that appending is 
allowed even if clobber=>0.

=item col_names

Explicitly specify the column names and order in which they are
written.  No checks for consistency with supplied data hash.

    col_names => [col_name1, col_name2, ...],

=item colkeys

Explicitly specify the column information using the column keys
information returned by fits_read_bintbl.  This provides a 
convenient way to duplicate the columns from an input file.  
See the examples below.  If colkeys is used for this purpose, 
it must be updated accordingly if the data column characteristics 
are changed (i.e. columns removed or added or changed in size).  

    colkeys => $colkeys,

=item append

Append the binary table to an existing file.  If no file exists, it
will be created.

=item close

Close the file after writing the binary table.  Defaults to 1.  
If set to 0, the additional data can be written to the table
by calling fits_write_bintbl again with the FITS file pointer
in place of the file name.

=item keywords
	
Specify a set of header keywords to be included in the extension.
The keywords are passed as a list reference.  Each list element
can either be a list or hash reference, as illustrated below:
  
    keywords => [[<keyword>, <value>, <type>, <comment>],
                 [...],
                 {key => <keyword>, val => <value>, 
                  type => <type>, comment => <comment>},
                 ...
                ];

The keyword <type> and <comment> are optional.  If <type> is not
specified, then 'STRING' is used.  The possible user-specified types are
'DOUBLE', 'FLOAT', 'SHORT', 'INT', 'LONG', 'BYTE', 'USHORT', 'ULONG',
and 'STRING'.  (Case is not important).

=item primary_hdu

Specify a FITS file name from which to copy the primary HDU header keywords.
  
    primary_hdu => <fits_file>

=item tform

Specify the data type for each non-PDL column of the input data hash.
This is used to define the FITS TFORMn keyword.
If not supplied, fits_write_bintbl() will make a reasonable guess.
Tform is a hash reference, keyed off the column names.  

    tform => { <name1> => '<repeat1><type1>',
	       <name2> => '<repeat2><type2>',
	       ... },

The <type> can be any of the types listed above for the C<keywords> 
option.  The <repeat> count should only be different from "1" for STRING columns, 
in which case <repeat> implies the maximum string length.
Note that the <type> here can also be a valid FITS data type, e.g.
'1J' for a long integer, '1E' for a float.

=item tunit

Specify the units for each column of the input data hash.
This is used to define the FITS TUNITn keyword.

    tunit => { <name1> => '<units1>',
	       <name2> => '<units2>',
	       ... },

=head1 EXAMPLES
 # Read all the data in the first binary table extension of file, and write
 # back to a new file.  This makes a nearly identical copy of file.

 ($colkeys, %data) = fits_read_bintbl("file.fits", {colkeys=>1}); 
 $hdr = fits_read_hdr("file.fits", {list=>1});
 fits_write_bintbl('new_file.fits', %data, {colkeys=>$colkeys,
                                            keywords=>$hdr}, 
                                            primary_hdu=>'file.fits'});

 # Read the list of columns from the first extension of 'fidpr.fits'
 
 %data = fits_read_bintbl("fidpr.fits", 'p_lsi', 'slot', 'id_string'); 
 
 # Read the columns from the named extension
 
 %data = fits_read_bintbl("fidpr.fits", 'p_lsi', 'slot', 'mag_i_avg', 'id_string',
			     {extname => 'FIDPROPS'}); 
 
 # Minimal statement to read the first bin table in file
 
 %data = fits_read_bintbl("fidpr.fits");
 
 # Now write the data back to a file.  By default, if the file exists then
 # it will be over-written
 
 $ok = fits_write_bintbl("test_out1.fits", %data);

 # Write a new file using header keywords from a template file.  

 $hdr = fits_read_hdr("fidpr.fits", 2);  # read keywords from 2nd HDU
 $ok = fits_write_bintbl("test_out1.fits", %data, {keywords => $hdr,
                                                   primary_hdu => "fidpr.fits.fits"});

 # Add another extension, this time explicitly specifying the
 # size of the string columns

 $ok = fits_write_bintbl("test_out1.fits", %data, 
   { clobber => 1, # This is the DEFAULT
     append  => 1,
     tform => { id_string => '9A', # 9 character string
		id_status => '4 string', # 4 char. string
              } 
   } );

 # A complicated example, with mixed Perl lists and Piddles

 $ok = fits_write_bintbl("test_out2.fits", %data2,
     { extname => 'BINTABLE2',

       # Don't overwrite an existing file
       clobber => 0,

       # Set some header keywords.  The order is [name, val, type, comment].
       # Comment and type are optional.  If no type is supplied, 'string' is used.
       keywords => [['testkey1', 1.2345e2, 'double', 'Test double keyword'],
     	            ['testkey2', 1.2345e2, 'float'], 
     	       	    ['testkey3', 1.2345e2],  
     	       	    ['testkey4', 12, 'long', 'Test keyword long']],

       # Explicitly specify some of the data types.  This is not required,
       # but useful for efficiency or if you want something besides LONG,
       # DOUBLE, or STRING.
       tform => { intarr => 'ushort',
       	          strarr => '10 A',
     	          fltarr => 'float'},
       
       # Set the units
       tunit => { fltarr => 'gm/cm^2',
     	          intarr => 'meters' },
       
   } );

=head1 LIMITATIONS

Current code cannot write a multidimensional Perl Array.  Method for copying
header keywords from one HDU into another is hacked up and may not be correct
under all circumstances.

=head1 BUGS

Probably.

=head1 AUTHOR

Tom Aldcroft (taldcroft@cfa.harvard.edu)

=cut
