package Ska::Run;

use 5.008;
use strict;
use warnings;
use Ska::Message;
use Carp;

require Exporter;

our @ISA = qw(Exporter);

# Items to export into callers namespace by default. Note: do not export
# names by default without a very good reason. Use EXPORT_OK instead.
# Do not simply export all your public functions/methods/constants.

# This allows declaration	use Ska::Run ':all';
# If you do not need this, moving things directly into @EXPORT or @EXPORT_OK
# will save memory.
our %EXPORT_TAGS = ( 'all' => [ qw(
				   run
				  ) ] );

our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );

our @EXPORT = qw(
		 run
		);

our $VERSION = '0.04';

our $LOUD = 0;
our $MESSAGE;
our $TIMEOUT;
our $CROAK;


# Preloaded methods go here.
##***************************************************************************
sub run {
##***************************************************************************
    my $cmds = shift;
    my $cmd_pid;
    my $cmd_root;
    my @out;
    my $status;
    local $_;

    # Set up run parameters, including optional params spec'd after cmd
    my %par = (loud    => $LOUD,
	       message => $MESSAGE,
	       timeout => $TIMEOUT,
	       croak   => $CROAK,
	       @_
	      );

    my $msg = $par{message} || Ska::Message->new( stdout => $par{loud} );

    # Run within eval block to be able to set a local alarm to time out if
    # command doesn't finish in in $par{timeout} seconds
    # See perldoc -f alarm
    eval {
	local $SIG{ALRM} = sub { die "alarm\n";
			       }; # NB: \n required

	alarm $par{timeout} if $par{timeout};

	# Set up commands to run.  Arg to run() can be either a command or an
	# reference to a list of commands
	my @cmds = (ref $cmds eq "ARRAY") ? @{$cmds} : ($cmds);
	for my $cmd (@cmds) {
	    ($cmd_root) = split ' ', $cmd;
	    chomp $cmd;		# Fix heredoc or string with trailing \n
	    $msg->message("Running $cmd");
	    $cmd_pid = open CMD, "$cmd 2>&1 |" or die "ERROR - Could not start $cmd_root command\n";
	
	    while (<CMD>) {
		$msg->message($_);
		push @out, $_;
	    }
	    # Close and make sure all is OK.  'die $! ? "A" : "B"' doesn't parse in xemacs)
	    unless (close CMD) {
		if ($!) { die "ERROR - Couldn't close $cmd_root pipe: $! \n"; }
		else    { die "WARNING - '$cmd_root' returned non-zero status: $?\n"; }
	    }
	}

	alarm 0;
    };

    if ($@) {
	unless ($@ eq "alarm\n") {   # propagate unexpected errors
	    croak $@ if $par{croak};
	    return (wantarray ? ($@, @out) : $@);
	}

	my $warning = "WARNING - $cmd_root command timed out ".localtime()."\n";
	$status = $warning;
	$msg->message($warning);
				 
	# Kill the cmd process.  See Ska::Process for a more detailed routine that
	# finds all child processes with same group pid and kills them one at a time.
	# This is a bad idea here because there may be other "cron" processes with
	# same gid that should not be killed

	kill 9 => $cmd_pid;
	sleep 3;
	close CMD;
    }

    croak $status if ($status and $par{croak});
    return wantarray ? ($status, @out) : $status; # For this routine, undef => success
}

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

Ska::Run - Run a system command

=head1 SYNOPSIS

  use Ska::Run;

  ($status, @output) = run("ls -l", loud => 1);
  die $status if defined $status;

  my $msg = new Ska::Message(stdout => 1, time => 1);
  $status = run('echo exit | idl', message => $msg);
  die $status if defined $status;

  # Make run loud by default
  $Ska::Run::LOUD = 1;

  # This command will time out
  run("sleep 20", timeout => 5);  

=head1 DESCRIPTION

Ska::Run provides a clean way of running system commands, capturing the output,
and ensuring a timely exit.

=over 8

=item B<run(command, option =E<gt> val, ...)>

If called in a scalar context, run() returns only the status, which is
undefined if the command ran to completion, and otherwise contains an error
message.  The Unix exit status of the command is tested (via the pipe close)
and a non-zero value will return a warning in the return status.  If called in
an array context, run() returns both the status and an array of the command
output.

=head2 Options

=over 8

=item B<loud>

Automatically print STDOUT from command.  Default = undef

=item B<message>

Send output to the supplied Ska::Message object

=item B<timeout>

Return control to the calling program if command does not finish within the specified
duration (seconds).   

=back

=back

=head2 Configuration

Each of the above options has an associated package variable (in all capital letters) 
that defines the default for each call to run().  E.g.

 $Ska::Run::TIMEOUT = 1500;   # Set timeout for all run() calls to 1500 seconds

=head1 AUTHOR

Tom Aldcroft, E<lt>aldcroft@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2005 by Tom Aldcroft

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.3 or,
at your option, any later version of Perl 5 you may have available.


=cut
